import logging
import os
import typing
from json import JSONDecodeError
from typing import Optional, Type

import orjson
from loguru import logger
from postgrest.exceptions import APIError, generate_default_error_message
from pydantic import BaseModel
from supabase._async.client import AsyncClient
from supabase._async.client import create_client as async_create_client
from supabase.client import Client, ClientOptions, create_client
from tenacity import after_log, retry, stop_after_attempt, wait_exponential

from good_common.dependencies import AsyncBaseProvider, BaseProvider

from ._auth import GoogleAuthCallbackHandler

logging.getLogger("httpx").setLevel(logging.WARNING)

InsertableObject = dict | BaseModel
Insertable = InsertableObject | list[InsertableObject]


class SupabaseAsync:
    def __init__(
        self,
        client: AsyncClient,
        log_requests: bool = False,
    ):
        self._client = client

    @classmethod
    async def create(
        cls,
        url: str,
        key: str,
        schema: str = "public",
    ):
        client = await async_create_client(
            supabase_url=url,
            supabase_key=key,
            options=ClientOptions(
                schema=schema,
                postgrest_client_timeout=180,
            ),
        )
        return cls(client)

    @property
    def client(self) -> AsyncClient:
        return self._client

    def _preprocess_record(self, record: InsertableObject):
        if isinstance(record, BaseModel):
            return record.model_dump(mode="json", exclude_none=True)
        return record

    def _preprocess_data(self, data: Insertable):
        if not isinstance(data, list):
            data = [data]
        return [self._preprocess_record(record) for record in data]

    def _model_metadata(self, data: Insertable):
        if isinstance(data, list):
            _model = data[0]
        else:
            _model = data

        if isinstance(_model, BaseModel):
            model = _model.__class__
        elif isinstance(_model, dict):
            return {}

        fields = {}
        for name, field in model.model_fields.items():
            fields[name] = (field.json_schema_extra or {}).get("supabase", {})

        return {"model": model, "fields": fields}

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1.5, min=4, max=10),
        after=after_log(logger, logging.DEBUG),
        reraise=True,
    )
    async def upsert(
        self,
        table: str,
        data: Insertable,
        on_conflict: str | None = None,
        **kwargs,
    ):
        field_metadata = self._model_metadata(data)

        if on_conflict is None:
            _conflict_fields = []
            for field, metadata in field_metadata.get("fields").items():
                if metadata.get("primary_key") is True:
                    _conflict_fields.append(field)
                elif metadata.get("unique") is True:
                    _conflict_fields.append(field)
            if len(_conflict_fields) > 0:
                on_conflict = ",".join(_conflict_fields)
        response = None
        try:
            response = (
                await self.client.table(table)
                .upsert(
                    self._preprocess_data(data),
                    on_conflict=on_conflict,
                    **kwargs,
                )
                .execute()
            )
        except Exception as e:
            logger.error(e)
            if e.code == "21000":
                logger.info("inserting one by one")
                for row in data:
                    (
                        await self.client.table(table)
                        .upsert(
                            self._preprocess_data([row]),
                            on_conflict=on_conflict,
                            **kwargs,
                        )
                        .execute()
                    )

            # raise e
        if not response:
            return None
        if field_metadata.get("model"):
            output = [
                field_metadata.get("model").model_validate(record)
                for record in response.data
            ]
        else:
            output = response.data

        if isinstance(data, list):
            return output
        return output[0]

    async def fetch_all(
        self,
        table: str,
        columns: list[str] = ["*"],
        output_model: Type[BaseModel] = None,
    ):
        if output_model:
            columns = output_model.model_fields.keys()
        query = self.client.table(table).select(*columns)
        response = await query.execute()
        if output_model:
            return [output_model(**record) for record in response.data]
        return response.data

    async def fetch_filter(
        self,
        table: str,
        filter_func: typing.Callable,
        output_model: Type[BaseModel] | None = None,
    ):
        # columns = output_model.model_fields.keys()
        # logger.info(columns)
        response = await (filter_func(self.client.table(table).select("*"))).execute()
        if output_model:
            return [output_model(**record) for record in response.data]
        else:
            return response.data

    def google_auth(self, port: int = 9997):
        return GoogleAuthCallbackHandler(self.client, port)

    async def rpc(
        self,
        procedure: str,
        params: Optional[dict] = None,
        _cast_to: Optional[Type[BaseModel]] = None,
        _log: Optional[bool] = None,
        **_params,
    ):
        params = params or {}

        params.update(_params)

        _last_log = None
        _cast_to = _cast_to or params.pop("cast_to", None)
        _log = _log or params.pop("log", None)

        if _log is not None:
            _last_log = self._log_requests
            self._log_requests = _log

        query = self.client.postgrest.rpc(
            procedure,
            orjson.loads(
                orjson.dumps(params)
            ),  # ensures json properly serialized for requests
        )

        try:
            r = await query.execute()

        except Exception as e:
            raise APIError(str(e))

        try:
            return r
        except JSONDecodeError:
            raise APIError(generate_default_error_message(r))

        # if _log is not None:
        #     self._log_requests = _last_log

        # if data:
        #     if _cast_to and isinstance(data, dict):
        #         return _cast_to(**data)
        #     elif _cast_to and isinstance(data, list):
        #         return [_cast_to(**o) for o in data]
        #     return data
        return None


class Supabase:
    def __init__(
        self,
        client: Client,
        log_requests: bool = False,
    ):
        self._client = client
        self._log_requests = log_requests

    @property
    def client(self) -> Client:
        return self._client

    def log(self, msg: typing.Any):
        if self._log_requests:
            logger.info(msg)

    # @retry(
    #     stop=stop_after_attempt(2),
    #     wait=wait_exponential(multiplier=1, min=4, max=10)
    # )
    def rpc(
        self,
        procedure: str,
        params: Optional[dict] = None,
        _cast_to: Optional[Type[BaseModel]] = None,
        _log: Optional[bool] = None,
        **_params,
    ) -> typing.Any:
        params = params or {}

        params.update(_params)

        _last_log = None
        _cast_to = _cast_to or params.pop("cast_to", None)
        _log = _log or params.pop("log", None)

        if _log is not None:
            _last_log = self._log_requests
            self._log_requests = _log

        query = self.client.postgrest.rpc(
            procedure,
            orjson.loads(
                orjson.dumps(params)
            ),  # ensures json properly serialized for requests
        )

        try:
            r = query.execute()

        except Exception as e:
            raise APIError(str(e))

        try:
            return r
        except JSONDecodeError:
            raise APIError(generate_default_error_message(r))

        # if _log is not None:
        #     self._log_requests = _last_log

        # if data:
        #     if _cast_to and isinstance(data, dict):
        #         return _cast_to(**data)
        #     elif _cast_to and isinstance(data, list):
        #         return [_cast_to(**o) for o in data]
        #     return data
        # return None

    def _format_param_value(self, arg):
        # if isinstance(arg, dict):
        #     return orjson.dumps(arg).decode('utf-8')
        # if dataclasses.is_dataclass(arg):
        #     return orjson.dumps(dataclasses.asdict(arg)).decode('utf-8')
        return arg

    def _format_params(self, params: dict):
        return {
            k: self._format_param_value(v) for k, v in params.items() if v is not None
        }

    def google_auth(self, port: int = 9997):
        return GoogleAuthCallbackHandler(self.client, port)


class SupabaseProvider(BaseProvider[Supabase], Supabase):
    def _create_client(
        self,
        url: Optional[str] = None,
        key: Optional[str] = None,
        schema: Optional[str] = None,
        options: Optional[dict] = None,
    ):
        url = url or os.environ.get("SUPABASE_URL")
        key = key or os.environ.get("SUPABASE_KEY")

        if not url or not key:
            raise ValueError("Missing SUPABASE_URL or SUPABASE_KEY")

        if not options:
            options = {}
        if schema:
            options["schema"] = schema

        supabase: Client = create_client(
            url, key, options=ClientOptions(postgrest_client_timeout=180, **options)
        )
        return supabase

    def __init__(
        self,
        url: Optional[str] = None,
        key: Optional[str] = None,
        schema: Optional[str] = None,
        **kwargs,
    ):
        kwargs["url"] = url
        kwargs["key"] = key
        kwargs["schema"] = schema
        super().__init__(**kwargs)

    def initializer(
        self,
        cls_args: typing.Tuple[typing.Any, ...],
        cls_kwargs: typing.Dict[str, typing.Any],
        fn_kwargs: typing.Dict[str, typing.Any],
    ):
        # print((cls_args, cls_kwargs, fn_kwargs))
        return cls_args, {
            "client": self._create_client(
                **{
                    **cls_kwargs,
                    **{k: v for k, v in fn_kwargs.items() if k in ["schema"]},
                }
            )
        }


class SupabaseAsyncProvider(AsyncBaseProvider[SupabaseAsync], SupabaseAsync):
    def __init__(
        self,
        url: Optional[str] = None,
        key: Optional[str] = None,
        schema: Optional[str] = None,
        **kwargs,
    ):
        kwargs["url"] = url or os.environ.get("SUPABASE_URL")
        kwargs["key"] = key or os.environ.get("SUPABASE_KEY")
        kwargs["schema"] = schema or os.environ.get("SUPABASE_SCHEMA", "public")
        super().__init__(**kwargs)

    @classmethod
    async def provide(cls, *args, **kwargs):
        return await SupabaseAsync.create(*args, **kwargs)



# import asyncio
# import os
# import typing

# from loguru import logger
# from pydantic import BaseModel, Field
# # from realtime.connection import Socket
# from realtime import AsyncRealtimeChannel, AsyncRealtimeClient
# # from realtime import AsyncRealtimeClient

# from good_common.dependencies import BaseProvider

# T = typing.TypeVar("T", bound=BaseModel | dict)


# class WebhookRecord(BaseModel, typing.Generic[T]):
#     id: int
#     created_at: str
#     domain: str
#     body: T | dict
#     event: str


# class ChannelMessage(BaseModel, typing.Generic[T]):
#     columns: typing.List[dict]
#     commit_timestamp: str
#     errors: typing.Optional[typing.Any]
#     record: WebhookRecord[T]
#     table_schema: str = Field(alias="schema")
#     table: str
#     type: str


# # class Channel:
# #     def __init__(self, socket: Socket, channel_name: str):
# #         self._socket = socket
# #         self._channel_name = channel_name
# #         self._queue = asyncio.Queue()
# #         self._listener = None
# #         self._stop = asyncio.Event()

# #     def channel_callback(self, message: dict):
# #         self._queue.put_nowait(message)

# #     async def __aenter__(self):
# #         await self._socket._connect()
# #         self._channel = self._socket.set_channel(self._channel_name)
# #         return self

# #     async def __aexit__(self, exc_type, exc_value, traceback):
# #         logger.info("Existing channel gracefully")
# #         self._stop.set()
# #         if self._listener:
# #             self._listener.cancel()

# #     async def listen(self):
# #         try:
# #             await asyncio.gather(self._socket._listen(), self._socket._keep_alive())
# #         except asyncio.CancelledError:
# #             pass
# #         except Exception as e:
# #             logger.info(("listen", e))
# #             self._stop.set()

# #     async def on(
# #         self,
# #         event: typing.Literal["insert", "update", "delete"],
# #         cast_as: typing.Type[T] | None = None,
# #     ) -> typing.AsyncGenerator[ChannelMessage[T], None]:
# #         try:
# #             await self._channel._join()  # type: ignore

# #             self._channel.on(event.upper(), self.channel_callback)  # type: ignore

# #             loop = asyncio.get_running_loop()

# #             self._listener = loop.create_task(self.listen())

# #             # loop.ad

# #             while not self._stop.is_set():
# #                 try:
# #                     message = await asyncio.wait_for(self._queue.get(), timeout=1)
# #                     # yield ChannelMessage[T](**message)
# #                     yield ChannelMessage[cast_as](**message) if cast_as else message
# #                 except asyncio.TimeoutError:
# #                     # continue  # Timeout reached, check the stop event again
# #                     continue
# #                 except asyncio.CancelledError:
# #                     self._stop.set()
# #                     break
# #                 except Exception as e:
# #                     logger.info(("on", e))
# #                     await asyncio.sleep(1)
# #                     return
# #         finally:
# #             self._listener.cancel()  # type: ignore
# #             # Ensure the listener task is cancelled when exiting
# #             try:
# #                 await self._listener  # type: ignore
# #                 # Wait for the listener task to be cancelled
# #             except asyncio.CancelledError:
# #                 pass

# #             return


# # class SupabaseRealtime:
# #     def __init__(
# #         self,
# #         supabase_id: str,
# #         supabase_key: str,
# #     ):
# #         self._supabase_id = supabase_id
# #         self._supabase_key = supabase_key

# #         self._socket = Socket(self._socket_url)

# #     @property
# #     def _socket_url(self) -> str:
# #         return (
# #             f"wss://{self._supabase_id}.supabase.co/realtime/v1"
# #             f"/websocket?apikey={self._supabase_key}&vsn=1.0.0"
# #         )

# #     def channel(self, channel_name: str):
# #         return Channel(self._socket, channel_name)

# #     async def __aenter__(self):
# #         return self

# #     async def __aexit__(self, exc_type, exc_value, traceback):
# #         logger.info(("aexit", exc_type, exc_value, traceback))
# #         await self._socket.close()


# # class SupabaseRealtimeProvider(BaseProvider[SupabaseRealtime], SupabaseRealtime):
# #     def __init__(
# #         self, supabase_id: str | None = None, supabase_key: str | None = None, **kwargs
# #     ):
# #         kwargs["supabase_id"] = supabase_id or os.environ["SUPABASE_ID"]
# #         kwargs["supabase_key"] = supabase_key or os.environ["SUPABASE_KEY"]
# #         super().__init__(**kwargs)

# #     def initializer(self, cls_args, cls_kwargs, fn_kwargs):
# #         cls_kwargs["supabase_id"] = fn_kwargs.get(
# #             "supabase_id", os.environ["SUPABASE_ID"]
# #         )
# #         cls_kwargs["supabase_key"] = fn_kwargs.get(
# #             "supabase_key", os.environ["SUPABASE_KEY"]
# #         )

# #         return cls_args, cls_kwargs



import typing
import http.server
import socketserver
from urllib.parse import urlparse, parse_qs
import threading
import functools
import time
import socket

class CallbackHandler(http.server.SimpleHTTPRequestHandler):
    def __init__(self, *args, success_callback: typing.Callable[[str], None], **kwargs):
        self.success_callback = success_callback
        super().__init__(*args, **kwargs)
    
    def do_GET(self):
        if self.path.startswith('/auth/callback'):
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            html = '''
            <html>
            <body>
                <script>
                    var hash = window.location.hash.substr(1);
                    var xhr = new XMLHttpRequest();
                    xhr.open('POST', '/token', true);
                    xhr.setRequestHeader('Content-Type', 'application/x-www-form-urlencoded');
                    xhr.onload = function() {
                        document.body.innerHTML += '<p>Authentication successful! You can close this window.</p>';
                    };
                    xhr.send(hash);
                </script>
                <p>Processing authentication...</p>
            </body>
            </html>
            '''
            self.wfile.write(html.encode())
        else:
            super().do_GET()

    def do_POST(self):
        if self.path == '/token':
            content_length = int(self.headers['Content-Length'])
            post_data = self.rfile.read(content_length).decode('utf-8')
            self.send_response(200)
            self.send_header('Content-type', 'text/plain')
            self.end_headers()
            self.wfile.write(b"Token received")
            self.success_callback(post_data)

class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    allow_reuse_address = True

    def server_bind(self):
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.server_address)

class GoogleAuthCallbackHandler:
    def __init__(self, supabase, port: int = 9997):
        self.supabase = supabase
        self.response = None
        self.port = port
        self.httpd = None
        self.server_thread = None
        self.auth_event = threading.Event()
        
    def run_server(self):
        while True:
            try:
                self.httpd = ThreadedTCPServer(("", self.port), functools.partial(CallbackHandler, success_callback=self.handle_callback))
                print(f"Serving at port {self.port}")
                self.httpd.serve_forever()
            except OSError as e:
                if e.errno == 48:  # Address already in use
                    print(f"Port {self.port} is busy, retrying in 1 second...")
                    time.sleep(1)
                else:
                    raise
            else:
                break

    def __enter__(self):
        self.server_thread = threading.Thread(target=self.run_server)
        self.server_thread.start()
        self.response = self.supabase.auth.sign_in_with_oauth({
            "provider": 'google',
            "options": {
                "redirect_to": f'http://localhost:{self.port}/auth/callback',
                'grant_type': 'id_token'
            }
        })
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.auth_event.wait()  # Wait for authentication to complete
        if self.httpd:
            self.httpd.shutdown()
            self.httpd.server_close()
        if self.server_thread:
            self.server_thread.join()

    def handle_callback(self, post_data):
        query_params = parse_qs(post_data.strip('#'))
        try:
            self.supabase.auth.set_session(query_params['access_token'][0], query_params['refresh_token'][0])
            # print(r)
        finally:
            self.auth_event.set()  # Signal that authentication is complete
